#
# mtof.rb
#
#   Copyright (c) 2001 Minero Aoki <aamine@loveruby.net>
#
#   This program is free software.
#   You can distribute/modify this program under the terms of
#   the GNU Lesser General Public License version 2 or later.
#

module ReFe

  class MtoF

    def self.build( f, db )
      db.mtof_updating {
        new.build f, db
      }
    end

    def build( f, db )
      while line = f.gets do
        if /\AInit_/ === line then
          _build f, db
        end
      end
    end

    def _build( f, db )
      while line = f.gets do
        case line
        when /rb_define_(?:singleton_)?method/
          until paren_match? line do
            line << f.gets.strip
          end
          db.register_mf_relation get_method_name(line),
                                  get_function_name(line)
        when /\A\}/
          return
        end
      end
    end

    def paren_match?( line )
      line.count('(') == line.count(')')
    end

    def get_method_name( line )
      getclass(line) + getmethod(line)
    end

    def getclass( line )
      v2c( getcvar(line) + classtype(line) )
    end

    def getcvar( line )
      m = /\((\w+)/.match( line ) or
              raise ArgumentError, "cannot find class name: #{line}"
      m[1]
    end

    def classtype( line )
      /define_singleton/ === line ? '.' : '#'
    end

    VarToClass = {
      'argf.'          => 'ARGF#',
      'envtbl.'        => 'ENV#',
      'ruby_top_self.' => 'main#',
      'cThGroup#'      => 'ThreadGroup#',
      'cThGroup.'      => 'ThreadGroup.'
    }

    def v2c( v )
      if v == 'rb_mKernel#' then
        'Object#'
      elsif /\Arb_[cme][A-Z]/ === v then
        v[4..-1]
      else
        VarToClass[v]
      end
    end

    def getmethod( line )
      m = /"((?:[^"\\]+|\\.)*)"/.match( line ) or
              raise ArgumentError, "cannot find method name: #{line}"
      m[1]
    end

    def get_function_name( line )
      m = /_method\(\w+,\s*"[^"\\]+",\s*(\w+),/.match( line ) or
              raise ArgumentError, "cannot find function name: #{line}"
      m[1]
    end
  
  end

end
